/*
 * s2kfix.c
 *
 * original code: SASANO Takayoshi
 * --- public domain, no warranty.
 *
 * ԻȤ
 *
 * s2kfix	Savage2000Υꥯåbiosp->videomaxclkƤ
 *		ꥯåͽ¬biosp->videomaxclk
 *		褢٤ͤɽ
 * s2kfix fix	biosp->videomaxclkƤ򡢥ꥯå򸵤˽
 *
 * ǤǻȤäƲ
 *   ΥץƤ˴ؤѡʥǥҤ
 *   䤤碌褦ʤȤϤʤǲǤޤΤǡˡ
 *
 *   ʤBBSǡ֥ե꡼ˤϥС󥢥åפԤǤʤפȤ
 *   ֳȯʤ鳫ȯĶget٤פȤȯͤ
 *   ΥɵڤӡɤƤʪλѤػߤޤ
 */

#include <basic.h>
#include <bstdio.h>
#include <tstring.h>
#include <driver/pcat/sys.h>
#include <driver/pcat/pci.h>
#include <driver/pcat/biosp.h>

#define	VENDOR_S3	0x5333
#define	DEVICE_S2K	0x9102

#define	F_REF		14318181	/* reference clock (Hz) */
#define	MEMORY_WIDTH	(64 / 8)	/* memory width (byte) */
#define	RAMDAC_MAX	350000000	/* RAMDAC max (Hz) */

#define	VGA_SRX		0x3c4

/* read from sequencer */
Inline	UB	ReadSR(UB idx)
{
	out_b(VGA_SRX, idx);
	return in_b(VGA_SRX + 1);
}

/* write to sequencer */
Inline	void	WriteSR(UB idx, UB data)
{
	out_b(VGA_SRX, idx);
	out_b(VGA_SRX + 1, data);
	return;
}

/* calc memory clock */
LOCAL	UW	get_memclk(void)
{
	UW	sr8, sr10, sr11, m, n, r, mclk;

	/* Unlock extended sequencer (SR8) */
	sr8 = ReadSR(0x08);
	WriteSR(0x08, 0x06);

	/* get MCLK value low(SR10) and high(SR11) */
	sr10 = ReadSR(0x10);
	sr11 = ReadSR(0x11);

	/* Lock extended sequencer */
	WriteSR(0x08, sr8);

	/* get M, N, R value */
	m = sr11 & 0x7f;
	n = sr10 & 0x1f;
	r = (sr10 >> 5) & 0x03;

	mclk = (F_REF * (m + 2)) / ((n + 2) * (1 << r));

	return mclk;
}

/* main */
EXPORT	ERR	main(W argc, TC *argv[])
{
	ERR	err;
	W	av[L_DEVCONF_VAL];
	UW	memclk, newclk;
	LOCAL	UB	pixbyte[] = {1, 2, 4};
	LOCAL	TC	key[] = {L""};

	/* check savage2k */
	if (searchPciDev(VENDOR_S3, DEVICE_S2K) < 0) {
		printf("Savage2000 not found\n");
		err = ER_NODEV;
		goto fin;
	}

	/* check videomode */
	if (GetDevConf("VIDEOMODE", av) < 1) {
		printf("VIDEOMODE undefined\n");
		err = ER_PAR;
		goto fin;
	} else if (av[0] < 1 || av[0] > 18) {
		printf("unsupported VIDEOMODE (%d)\n", av[0]);
		err = ER_NOSPT;
		goto fin;
	}	

	/* calc new videomaxclk */
	memclk = get_memclk();
	newclk = memclk * MEMORY_WIDTH / pixbyte[(av[0] - 1) % 3];
	if (newclk > RAMDAC_MAX) newclk = RAMDAC_MAX;

	/* display status */
	printf("Memory clock      : %dHz\n", memclk);
	printf("biosp->videomaxclk: %dHz\n", biosp->videomaxclk);
	printf("new videomaxclk   : %dHz\n", newclk);

	/* update */
	if (argc > 1 && !tc_strcmp(argv[1], key)) {
		biosp->videomaxclk = newclk;
		printf("biosp->videomaxclk updated\n");
	}

	err = ER_OK;
 fin:
	return err;
}
